@@ -1,69 +1,12 @@
 import re
+from collections import OrderedDict
 
 import pandocfilters as pf
 
-from attributes import PandocAttributes
+from pandocattributes import PandocAttributes
 
-
-figure_styles = {'latex': ('\n'
-                           '\\begin{{figure}}[htbp]\n'
-                           '\\centering\n'
-                           '\\includegraphics{{{filename}}}\n'
-                           '\\caption{star}{{{caption}}}\n'
-                           '\\label{{{id}}}\n'
-                           '\\end{{figure}}\n'),
-
-                 'html': ('\n'
-                          '<div {attrs}>\n'
-                          '<img src="{filename}" alt="{alt}" />'
-                          '<p class="caption">{fcaption}</p>\n'
-                          '</div>\n'),
-
-                 'html5': ('\n'
-                           '<figure {attrs}>\n'
-                           '<img src="{filename}" alt="{alt}" />\n'
-                           '<figcaption>{fcaption}</figcaption>\n'
-                           '</figure>'),
-
-                 'markdown': ('\n'
-                              '<div {attrs}>\n'
-                              '![{fcaption}]({filename})\n'
-                              '\n'
-                              '</div>\n')
-                 }
-
-# replacement text to use for in text internal links
-# that refer to various types of thing, in different
-# output formats
-latex_link = '{pre}\\autoref{{{label}}}{post}'
-html_link = '{pre}<a href="#{label}">{text}</a>{post}'
-markdown_link = '{pre}[{text}](#{label}){post}'
-
-latex_math_link = '{pre}\\autoref{{{label}}}{post}'
-html_math_link = '{pre}Equation \\eqref{{{label}}}{post}'
-markdown_math_link = '{pre}Equation $\\eqref{{{label}}}${post}'
-
-link_styles = {
-    'latex': {'figure': latex_link,
-              'section': latex_link,
-              'math': latex_math_link},
-    'html': {'figure': html_link,
-             'section': html_link,
-             'math': html_math_link},
-    'html5': {'figure': html_link,
-              'section': html_link,
-              'math': html_math_link},
-    'markdown': {'figure': markdown_link,
-                 'section': markdown_link,
-                 'math': markdown_math_link}
-}
-
-
-# http://cdn.mathjax.org/mathjax/latest/test/sample-eqrefs.html
-# mathjax claims that you can get equation references on regular
-# display math but I'm only finding it to work on equation
-# environments.
-# No, you can, just need autoequations: 'all' rather than 'AMS'
+from identity import *
+from styles import *
 
 
 def RawInline(format, string):
@@ -72,71 +15,63 @@ def RawInline(format, string):
     """
     if format == 'html5':
         format = 'html'
-        # pass
     return pf.RawInline(format, string)
 
 
-def rawlatex(s):
-    return pf.RawInline('latex', s)
+def RawBlock(format, string):
+    """Overwrite pandocfilters RawBlock so that html5
+    and html raw output both use the html writer.
+    """
+    if format == 'html5':
+        format = 'html'
+    return pf.RawBlock(format, string)
 
 
-def rawhtml(s):
-    return pf.RawInline('html', s)
-
-
-def rawmarkdown(s):
-    return pf.RawInline('markdown', s)
-
-
-def isheader(key, value):
-    return (key == 'Header')
-
-
-def islabeledmath(key, value):
-    return (key == 'Math' and re.search(r'\\label{\S*}', value[1]))
-
-
-# pattern that matches #reflink
-# only allow characters that we can have in latex labels
-# currently have to be after whitespace
-# terminated by whitespace, period or backslash
-imp_reflink_pattern = re.compile(r'([\s]?)(#[\w:&^]+)([\. \t\\]?)')
-
-
-def isinternalref(key, value):
-    # This can fall over if we don't create_figures from our
-    # special attr images first - it can match #id in the attrs
-    return key == 'Str' and imp_reflink_pattern.match(value)
-
 # define a new type for internal references [pre, label, post]
 InternalRef = pf.elt('InternalRef', 3)
-
-
-def isattr(string):
-    return string.startswith('{') and string.endswith('}')
-
-
-def isfigure(key, value):
-    return (key == 'Para' and len(value) == 2 and value[0]['t'] == 'Image')
-
-
-def isattrfigure(key, value):
-    return (key == 'Para'
-            and value[0]['t'] == 'Image'
-            and isattr(pf.stringify(value[1:])))
-
-
-def isdivfigure(key, value):
-    """Matches images contained in a Div with 'figure' as a class."""
-    return (key == 'Div' and 'figure' in value[0][1])
-
-
+# and multiple references [pre, [label], post]
+# https://tex.stackexchange.com/questions/15728/multiple-references-with-autoref
+# https://github.com/mathjax/MathJax/issues/71
+# http://docs.mathjax.org/en/latest/tex.html#automatic-equation-numbering
+MultiInternalRef = pf.elt('MultiInternalRef', 3)
 # define a new Figure type - an image with attributes
 Figure = pf.elt('Figure', 3)  # caption, target, attrs
 
 
-def isFigure(key, value):
-    return key == 'Figure'
+def create_pandoc_multilink(strings, refs):
+    inlines = [[pf.Str(str(s))] for s in strings]
+    targets = [(r, "") for r in refs]
+    links = [pf.Link(inline, target)
+             for inline, target in zip(inlines, targets)]
+
+    return join_items(links)
+
+
+def join_items(items, method='append'):
+    """Join the list of items together in the format
+
+    'item[0]' if len(items) == 1
+    'item[0] and item[1]' if len(items) == 2
+    'item[0], item[1] and item[2]' if len(items) == 3
+
+    and so on.
+    """
+    out = []
+    join_to_out = getattr(out, method)
+
+    join_to_out(items[0])
+
+    if len(items) == 1:
+        return out
+
+    for item in items[1: -1]:
+        out.append(pf.Str(', '))
+        join_to_out(item)
+
+    out.append(pf.Str(' and '))
+    join_to_out(items[-1])
+
+    return out
 
 
 def create_figures(key, value, format, metadata):
@@ -180,12 +115,16 @@ class ReferenceManager(object):
     section_count = [0, 0, 0, 0, 0, 0]
     figure_count = 0
     equation_count = 0
-    refdict = {}
+    references = {}
 
     replacements = {'figure': 'Figure {}',
                     'section': 'Section {}',
                     'math': 'Equation {}'}
 
+    multi_replacements = {'figure': 'Figures ',
+                          'section': 'Sections ',
+                          'math': 'Equations '}
+
     formats = ('html', 'html5', 'markdown', 'latex')
 
     def increment_section_count(self, header_level):
@@ -209,10 +148,21 @@ class ReferenceManager(object):
         """
         return '.'.join(str(i) for i in self.section_count[:header_level])
 
-    def consume_and_replace_references(self, key, value, format, metadata):
+    def consume_references(self, key, value, format, metadata):
+        """Find all figures, sections and math in the document
+        and append reference information to the reference state.
+        """
+        if isFigure(key, value):
+            self.consume_figure(key, value, format, metadata)
+        elif isheader(key, value) and format in self.formats:
+            self.consume_section(key, value, format, metadata)
+        elif islabeledmath(key, value):
+            self.consume_math(key, value, format, metadata)
+
+    def replace_references(self, key, value, format, metadata):
         """Find all figures, sections and equations that can be
-        referenced in the document and replace them with appropriate
-        text whilst appending to the internal refdict.
+        referenced in the document and replace them with format
+        appropriate substitutions.
         """
         if isFigure(key, value):
             return self.figure_replacement(key, value, format, metadata)
@@ -221,49 +171,87 @@ class ReferenceManager(object):
         elif islabeledmath(key, value):
             return self.math_replacement(key, value, format, metadata)
 
+    def consume_figure(self, key, value, format, metadata):
+        """If the key, value represents a figure, append reference
+        data to internal state.
+        """
+        _caption, (filename, target), (id, classes, kvs) = value
+        if 'unnumbered' in classes:
+            return
+        else:
+            self.figure_count += 1
+            self.references[id] = {'type': 'figure',
+                                   'id': self.figure_count,
+                                   'label': id}
+
+    def consume_section(self, key, value, format, metadata):
+        """If the key, value represents a section, append reference
+        data to internal state.
+        """
+        level, attr, text = value
+        label, classes, kvs = attr
+
+        if 'unnumbered' in classes:
+            return
+        else:
+            self.increment_section_count(level)
+            secn = self.format_section_count(level)
+            self.references[label] = {'type': 'section',
+                                      'id': secn,
+                                      'label': label}
+
+    def consume_math(self, key, value, format, metadata):
+        """If the key, value represents math, append reference
+        data to internal state.
+        """
+        self.equation_count += 1
+        mathtype, math = value
+        label, = re.search(r'\\label{([\w:&^]+)}', math).groups()
+        self.references[label] = {'type': 'math',
+                                  'id': self.equation_count,
+                                  'label': label}
+
     def figure_replacement(self, key, value, format, metadata):
-        """Replace figures with appropriate representation and
-        append info to the refdict.
+        """Replace figures with appropriate representation.
 
         This works with Figure, which is our special type for images
         with attributes. This allows us to set an id in the attributes.
 
         The other way of doing it would be to pull out a '\label{(.*)}'
-        from the caption of an Image and use that to update the refdict.
+        from the caption of an Image and use that to update the references.
         """
-        _caption, (filename, target), (id, classes, kvs) = value
+        _caption, (filename, target), attrs = value
         caption = pf.stringify(_caption)
 
-        if 'unnumbered' in classes:
+        attr = PandocAttributes(attrs)
+
+        if 'unnumbered' in attr.classes:
             star = '*'
             fcaption = caption
         else:
-            self.figure_count += 1
+            ref = self.references[attr.id]
             star = ''
-            fcaption = 'Figure {n}: {caption}'.format(n=self.figure_count,
-                                                      caption=caption)
-            self.refdict[id] = {'type': 'figure',
-                                'id': self.figure_count}
+            if caption:
+                fcaption = 'Figure {n}: {caption}'.format(n=ref['id'],
+                                                          caption=caption)
+            else:
+                fcaption = 'Figure {n}'.format(n=ref['id'])
 
-        if 'figure' not in classes:
-            classes.insert(0, 'figure')
-
-        attr = PandocAttributes((id, classes, kvs))
-        attrs = attr.to_html()
+        if 'figure' not in attr.classes:
+            attr.classes.insert(0, 'figure')
 
         if format in self.formats:
-            figure = figure_styles[format].format(attrs=attrs,
-                                                  id=id,
+            figure = figure_styles[format].format(attr=attr,
                                                   filename=filename,
                                                   alt=fcaption,
                                                   fcaption=fcaption,
                                                   caption=caption,
                                                   star=star)
-            return pf.Para([RawInline(format, figure)])
+
+            return RawBlock(format, figure)
 
     def section_replacement(self, key, value, format, metadata):
-        """Replace sections with appropriate representation and
-        append info to the refdict.
+        """Replace sections with appropriate representation.
         """
         level, attr, text = value
         label, classes, kvs = attr
@@ -271,11 +259,8 @@ class ReferenceManager(object):
         if 'unnumbered' in classes:
             pretext = ''
         else:
-            self.increment_section_count(level)
-            secn = self.format_section_count(level)
-            self.refdict[label] = {'type': 'section',
-                                   'id': secn}
-            pretext = '{}: '.format(secn)
+            ref = self.references[label]
+            pretext = '{}: '.format(ref['id'])
 
         pretext = [pf.Str(pretext)]
 
@@ -288,39 +273,45 @@ class ReferenceManager(object):
 
     def math_replacement(self, key, value, format, metadata):
         """Math should not need replacing as mathjax / latex will
-        take care of any references. All we do is append to the
-        refdict and increment an equation count (which nothing uses
-        yet).
+        take care of any references.
+
+        http://meta.math.stackexchange.com/questions/3764/equation-and-equation-is-the-same-for-me
         """
-        self.equation_count += 1
-        mathtype, math = value
-        label, = re.search(r'\\label{([\w:&^]+)}', math).groups()
-        self.refdict[label] = {'type': 'math',
-                               'id': self.equation_count}
-        return None
+        return
 
     def create_internal_refs(self, key, value, format, metadata):
         """Convert #label in the text into InternalRef, but only if the
-        label is in the refdict.
+        label is in the references.
         """
         if isinternalref(key, value):
             pre, link, post = imp_reflink_pattern.match(value).groups()
-            label = link.lstrip('#')
-            if label in self.refdict:
-                return InternalRef(pre, label, post)
+            labels = link.split('#')
+
+            # filter out labels not in references
+            labels = [label for label in labels if label in self.references]
+
+            if len(labels) == 0:
+                return None
+            elif len(labels) == 1:
+                return InternalRef(pre, labels[0], post)
+            elif len(labels) > 1:
+                return MultiInternalRef(pre, labels, post)
 
     def convert_internal_refs(self, key, value, format, metadata):
         """Convert all internal links from '#blah' into format
         specified in self.replacements.
         """
-        if key != 'InternalRef':
+        if key not in ('InternalRef' 'MultiInternalRef'):
             return None
 
+        elif key == 'MultiInternalRef':
+            return self.convert_multiref(key, value, format, metadata)
+
         else:
             pre, label, post = value
 
-        rtype = self.refdict[label]['type']
-        n = self.refdict[label]['id']
+        rtype = self.references[label]['type']
+        n = self.references[label]['id']
         text = self.replacements[rtype].format(n)
 
         if format in self.formats or True:
@@ -330,10 +321,48 @@ class ReferenceManager(object):
                                                      post=post)
             return RawInline(format, link)
 
+    def convert_multiref(self, key, value, format, metadata):
+        """Convert all internal links from '#blah' into format
+        specified in self.replacements.
+        """
+        if key != 'MultiInternalRef':
+            return
+
+        pre, labels, post = value
+
+        if format == 'latex':
+            link = latex_multi_link.format(pre=pre,
+                                           post=post,
+                                           labels=','.join(labels))
+            return RawInline('latex', link)
+
+        else:
+            D = [self.references[label] for label in labels]
+            # uniquely ordered types
+            types = list(OrderedDict.fromkeys(d['type'] for d in D))
+
+            links = []
+
+            for t in set(types):
+                n = [d['id'] for d in D if d['type'] == t]
+                labels = ['#' + d['label'] for d in D if d['type'] == t]
+                multi_link = create_pandoc_multilink(n, labels)
+
+                if len(labels) == 1:
+                    multi_link.insert(0,
+                                      pf.Str(self.replacements[t].format('')))
+                else:
+                    multi_link.insert(0, pf.Str(self.multi_replacements[t]))
+
+                links.append(multi_link)
+
+            return join_items(links, method='extend')
+
     @property
     def reference_filter(self):
         return [create_figures,
-                self.consume_and_replace_references,
+                self.consume_references,
+                self.replace_references,
                 self.create_internal_refs,
                 self.convert_internal_refs]
 
@@ -376,17 +405,10 @@ def toJSONFilter(actions):
     pf.json.dump(altered, pf.sys.stdout)
 
 
-def suppress_input_cells(key, value, format, metadata):
-    """For use with notedown. Suppresses code cells that have the
-    attribute '.input'.
-    """
-    if format == 'latex' and key == 'CodeBlock' and 'input' in value[0][1]:
-        return pf.Null()
-
-
 def main():
     refmanager = ReferenceManager()
-    toJSONFilter(refmanager.reference_filter + [suppress_input_cells])
+    toJSONFilter(refmanager.reference_filter)
+
 
 if __name__ == '__main__':
     main()
